#include "ellen/gif.h"
#include "gif/gif_lib.h"

const u16 gifIOff[] = { 0, 4, 2, 1 };
const u16 gifIJmp[] = { 8, 8, 4, 2 };

int gifReadFunc(GifFileType* gifInfo, GifByteType* buf, int count)
{
	return fread(buf, sizeof(GifByteType), count, (FILE*)gifInfo->UserData);
}

pGifHandler gifHandlerLoad(const char* filename)
{
	FILE* gifFile = fopen(filename, "rb");
	if(!gifFile)
		return NULL;
		
	GifFileType* gifInfo = DGifOpen((void*)gifFile, gifReadFunc);
	if(!gifInfo)
		return NULL;
		
	pGifHandler gifHandler = (pGifHandler)malloc(sizeof(tGifHandler));
	if(!gifHandler)
		return NULL;
		
	gifHandler->dimX = gifInfo->SWidth;
	gifHandler->dimY = gifInfo->SHeight;
		
		// Get GIF palette
	gifHandler->colCount = gifInfo->SColorMap->ColorCount;
	gifHandler->gifPalette = (u16*)malloc(gifHandler->colCount * PALETTE_BPP);
	for(u16 i=0; i<gifHandler->colCount; ++i) {
		GifColorType* pColor = &gifInfo->SColorMap->Colors[i];
		gifHandler->gifPalette[i] = RGB8(pColor->Red, pColor->Green, pColor->Blue);
	}
		
		// Pre-set variables
	gifHandler->loopCnt = 1;
	gifHandler->frmCount = 0;
	gifHandler->gifFrames = NULL;
	gifHandler->gifTiming = NULL;
	gifHandler->dataMode = GIFMODE_BITMAP;
		
		// Get GIF frames
	GifRecordType gifRecType;
	do {
		if(DGifGetRecordType(gifInfo, &gifRecType) == GIF_ERROR) {
			gifHandlerDestroy(gifHandler);
			return NULL;
		}
			
		switch(gifRecType) {
			case IMAGE_DESC_RECORD_TYPE: {
						
					if(DGifGetImageDesc(gifInfo) == GIF_ERROR) {
						gifHandlerDestroy(gifHandler);
						return NULL;
					}
						
					int gifWidth = gifInfo->Image.Width;
					int gifHeight = gifInfo->Image.Height;
					const u32 sizeofFrame = gifWidth * gifHeight;
						
					if(gifHandler->gifFrames == NULL)
						gifHandler->gifFrames = (u8*)malloc(sizeofFrame);
					else
						gifHandler->gifFrames = (u8*)realloc(gifHandler->gifFrames, (gifHandler->frmCount + 1) * sizeofFrame);
						
					u8* const curFrame = gifHandler->gifFrames + (sizeofFrame * gifHandler->frmCount);
					if (gifInfo->Image.Interlace) {
						for (int i=0; i<4; ++i) {
							for (int j=gifIOff[i]; j<gifHeight; j+=gifIJmp[i]) {
								if(DGifGetLine(gifInfo, curFrame + (i*gifWidth), gifWidth) == GIF_ERROR) {
									gifHandlerDestroy(gifHandler);
									return NULL;
								}
							}
						}
					} else {
						for (int i=0; i<gifHeight; ++i) {
							if(DGifGetLine(gifInfo, curFrame + (i*gifWidth), gifWidth) == GIF_ERROR) {
								gifHandlerDestroy(gifHandler);
								return NULL;
							}
						}
					}
						
						// Replaceing transparent color
					for(u32 i=0; i<sizeofFrame; ++i) {
						if(curFrame[i] == gifInfo->Image.transCol)
							curFrame[i] = 0;
						else if(curFrame[i] == 0)
							curFrame[i] = gifInfo->Image.transCol;
					}
						
						// Frame decoded
					gifHandler->frmCount += 1;
				} break;
			case EXTENSION_RECORD_TYPE: {
					GifByteType* gifExt;
				    int	gifExtCode;
						
					if(DGifGetExtension(gifInfo, &gifExtCode, &gifExt) == GIF_ERROR) {
						gifHandlerDestroy(gifHandler);
						return NULL;
					}
						
					switch(gifExtCode) {
						case APPLICATION_EXT_FUNC_CODE:
								if(memcmp(gifExt, "NETSCAPE2.0", 11))
									gifHandler->loopCnt = ((u16*)++gifExt)[16];
							break;
						case GRAPHICS_EXT_FUNC_CODE:
								gifInfo->Image.transCol = gifExt[4];
									
								if(!gifHandler->gifTiming)
									gifHandler->gifTiming = (u32*)malloc(sizeof(u32));
								else
									gifHandler->gifTiming = (u32*)realloc(gifHandler->gifTiming, sizeof(u32) * (gifHandler->frmCount + 1));
									
								gifHandler->gifTiming[gifHandler->frmCount] = ((u16*)gifExt)[1] * 10;
							break;
						default:
							break;
					}
						
					while(gifExt != NULL) {
						if(DGifGetExtensionNext(gifInfo, &gifExt) == GIF_ERROR) {
							gifHandlerDestroy(gifHandler);
							return NULL;
						}
					}
				} break;
			default:
				break;
		}
    } while (gifRecType != TERMINATE_RECORD_TYPE);
		
		// Swapping transparent color and pal0
	gifHandler->gifPalette[gifInfo->Image.transCol] ^= gifHandler->gifPalette[0];
	gifHandler->gifPalette[0] ^= gifHandler->gifPalette[gifInfo->Image.transCol];
	gifHandler->gifPalette[gifInfo->Image.transCol] ^= gifHandler->gifPalette[0];
		
	if(DGifCloseFile(gifInfo) == GIF_ERROR) {
		gifHandlerDestroy(gifHandler);
		return NULL;
	}
		
	fclose(gifFile);
		
	return gifHandler;
}
		
void gifHandlerDestroy(pGifHandler gifHandler)
{
		// Nothing to destroy / already destroyed
	if(gifHandler == NULL)
		return;
		
		// Destroy GIF frames
	free(gifHandler->gifFrames);
		// Destroy GIF palette
	free(gifHandler->gifPalette);
		// Destroy timing table
	free(gifHandler->gifTiming);
		// Destroy GIF handler
	free(gifHandler);
		
		// Mark as destroyed
	gifHandler = NULL;
}
		
bool gifConvertTiles(pGifHandler gifHandler)
{
		// Can't convert to tiles if width is not a multiple of 8px
	if(gifHandler->dimX & 0x07)
		return false;
		
		// Can't convert to tiles if height is not a multiple of 8px
	if(gifHandler->dimY & 0x07)
		return false;
		
		// Already tiled, return success :)
	if(gifHandler->dataMode == GIFMODE_TILES)
		return true;
		
		// Allocate new space for tiles
	u8* outTiles = (u8*)malloc(gifHandler->dimX * gifHandler->dimY * gifHandler->frmCount);
	if(!outTiles)
		return false;
		
		// Dirty bmp to tile conversion job
	u32 pX = 0, pY = 0;
	for(u32 curTile=0; curTile<(gifHandler->dimX * gifHandler->dimY * gifHandler->frmCount)/SIZEOF_TILE_256; ++curTile){
		for(u32 x=0; x<HW_TILE_WIDTH; ++x) {
			for(u32 y=0; y<HW_TILE_HEIGHT; ++y)
				outTiles[(curTile * SIZEOF_TILE_256) + x + (y * HW_TILE_HEIGHT)] = gifHandler->gifFrames[pX + x + ((pY + y) * gifHandler->dimX)];
		}
		pX += HW_TILE_WIDTH; 
		if (pX >= gifHandler->dimX) {
			pY += HW_TILE_HEIGHT;
			pX = 0;
		}
	}
		
		// Remove old bitmap data
	free(gifHandler->gifFrames);
		// Set it to be the new allocated tiles
	gifHandler->gifFrames = outTiles;
	gifHandler->dataMode = GIFMODE_TILES;
		
	return true;
}
